
/*

   flatland is a simple 2d physical simulator

   Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
   Written by Francois Fleuret <francois.fleuret@idiap.ch>

   This file is part of flatland

   flatland is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   flatland is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with flatland.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef POLYGON_H
#define POLYGON_H

#include "misc.h"
#include "canvas.h"

#ifdef XFIG_SUPPORT
#include "xfig_tracer.h"
#endif

#ifdef X11_SUPPORT
#include "simple_window.h"
#endif

class Polygon {
  struct Triangle {
    int a, b, c;
  };

  int _nb_max_polygons;
  int _nb_polygons; // We need to know the total number of polygons in
                    // the universe

  scalar_t _mass, _moment_of_inertia, _radius;

  scalar_t *_relative_x, *_relative_y;
  scalar_t _last_center_x, _last_center_y, _last_theta;

  int _total_nb_dots;
  int *_nb_dots;
  int *_effecting_edge;

  scalar_t *_length;
  Triangle *_triangles;

  inline void intersection(scalar_t xa2, scalar_t ya2,
                           scalar_t xa1, scalar_t ya1,
                           scalar_t xb2, scalar_t yb2,
                           scalar_t xb1, scalar_t yb1,
                           scalar_t &det, scalar_t &s1, scalar_t &s2) {
    scalar_t m11, m12, m21, m22, n11, n12, n21, n22, v1, v2;
    m11 = xa1 - xa2; m12 = xb2 - xb1; m21 = ya1 - ya2; m22 = yb2 - yb1;
    det = m11 * m22 - m12 * m21;
    if(det != 0) {
      n11 = + m22 / det; n12 = - m12 / det; n21 = - m21 / det; n22 = + m11 / det;
      v1 = xb2 - xa2; v2 = yb2 - ya2;
      s1 = n11 * v1 + n12 * v2; s2 = n21 * v1 + n22 * v2;
    }
  }

public:
  bool _initialized;
  bool _nailed;
  int _nb_vertices;
  scalar_t *_x, *_y;
  scalar_t _red, _green, _blue;
  scalar_t _center_x, _center_y, _theta;
  scalar_t _dcenter_x, _dcenter_y, _dtheta;

  // The x and y parameters are ignored if null. Useful to initialize
  // directly.

  Polygon(scalar_t mass,
          scalar_t red, scalar_t green, scalar_t blue,
          scalar_t *x, scalar_t *y,
          int nv);

  ~Polygon();

  Polygon *clone();

#ifdef XFIG_SUPPORT
  void color_xfig(XFigTracer *tracer);
  void print_xfig(XFigTracer *tracer);
#endif

#ifdef X11_SUPPORT
  void draw(SimpleWindow *window);
  void draw_contours(SimpleWindow *window);
#endif

  void draw(Canvas *canvas);
  void draw_contours(Canvas *canvas);

  void set_vertex(int k, scalar_t x, scalar_t y);
  void set_position(scalar_t center_x, scalar_t center_y, scalar_t theta);
  void set_speed(scalar_t dcenter_x, scalar_t dcenter_y, scalar_t dtheta);
  bool contain(scalar_t x, scalar_t y);
  void triangularize(int &nt, int nb, int *index);
  void initialize(int nb_polygons);
  bool update(scalar_t dt);
  scalar_t relative_x(scalar_t ax, scalar_t ay);
  scalar_t relative_y(scalar_t ax, scalar_t ay);
  scalar_t absolute_x(scalar_t rx, scalar_t ry);
  scalar_t absolute_y(scalar_t rx, scalar_t ry);

  void apply_force(scalar_t dt, scalar_t x, scalar_t y, scalar_t fx, scalar_t fy);
  void apply_border_forces(scalar_t dt, scalar_t xmin, scalar_t ymin, scalar_t xmax, scalar_t ymax);
  void apply_collision_forces(scalar_t dt, int n_polygon, Polygon *p);

  bool collide_with_borders(scalar_t xmin, scalar_t ymin,
                            scalar_t xmax, scalar_t ymax);

  bool collide(Polygon *p);
};

#endif
